//#include "mls.h"
#include "mlsfieldPaint.h"
#include <mlsfieldWindow.h>
#include <vector>
#include <fstream>
#include <time.h>
#include "vec2.h"

using namespace math;

ImagePaint::ImagePaint(QWidget *parent) : QGLWidget(parent) {
	m_spacePressed = false;
	m_mode = MODE_MOVE;
	m_toolMode = E_MOVE;
	m_index = 0;
	m_toolMode = 0;
	m_overlay = 0;
	m_zoom = m_resolution = 1;
	m_wndSize = QSize(0,0);
	m_imgLoaded = false;
	m_moveOriginKeep = QPointF(0.0,0.0);
	m_zoomOriginKeep = QPointF(0.0,0.0);

	m_isCPU=false;
	m_cpuRecompute=false;
	m_showFittingCircle = false;

	setAutoFillBackground(true);
	QPalette p = palette();
	p.setBrush(QPalette::Base, p.mid());
	setPalette(p);
	setCursor(Qt::OpenHandCursor);

	qApp->installEventFilter(this);
}


ImagePaint::~ImagePaint() {
}


void ImagePaint::restoreSettings(QSettings& settings) {
	setZoom(settings.value("zoom", 1.0).toDouble());
	setResolution(settings.value("resolution", 1.0).toDouble());
	setOriginX(settings.value("originX", 0).toDouble());
	setOriginY(settings.value("originY", 0).toDouble());
}


void ImagePaint::saveSettings(QSettings& settings) {
	settings.setValue("zoom", m_zoom);
	settings.setValue("resolution", m_resolution);
	settings.setValue("originX", m_moveOrigin.x());
	settings.setValue("originY", m_moveOrigin.y());
}


QPointF ImagePaint::view2image(const QPointF& p) const {
	QSize sz = size();
	float z = m_zoom /** m_resolution*/;
	QPointF res = QPointF(
		(p.x() - m_moveOrigin.x() * z - sz.width() / 2) / z + m_imgSize.width() / 2,
		(p.y() - m_moveOrigin.y() * z - sz.height() / 2) / z + m_imgSize.height() / 2	);
	return res;
}

QPointF ImagePaint::image2zoom(const QPointF& p) const {
	return p;
	//QSize sz = size();
	//float z = m_zoom /** m_resolution*/;
	//QPointF res = QPointF(
	//	(p.x() - m_origin.x() * z - sz.width() / 2) / z + m_imgSize.width() / 2,
	//	(p.y() - m_origin.y() * z - sz.height() / 2) / z + m_imgSize.height() / 2	);
	//return res;
}

QPointF ImagePaint::image2view(const QPointF& q) const {
	QSize sz = size();
	float z = m_zoom /** m_resolution*/;
	return QPointF(
		(q.x() - m_imgSize.width() / 2) * z + m_moveOrigin.x() * z + sz.width() / 2,
		(q.y() - m_imgSize.height() / 2) * z + m_moveOrigin.y() * z + sz.height() / 2 );
}

QTransform ImagePaint::viewTransform(const QSizeF& sz, double zoom, const QPointF& origin) const {
	QTransform tr;
	tr.translate(sz.width()/2.0f, sz.height()/2.0f);
	tr.translate(origin.x() * zoom, origin.y() * zoom);
	tr.scale(zoom, zoom);
	tr.translate(-imageSize().width()/2.0f, -imageSize().height()/2.0f);
	return tr;
}

float ImagePaint::pt2px(float pt) const {
	return pt / (m_zoom /** m_resolution*/);
}

void ImagePaint::setOverlay(QWidget *overlay) {
	if (overlay) {
		setMouseTracking(true);
		m_overlay = overlay;
		m_overlay->setVisible(false);
	} else {
		setMouseTracking(false);
		if (m_overlay) m_overlay->setVisible(false);
		m_overlay = 0;
	}
}

void ImagePaint::onIndexChanged(int idx)
{
	m_index = idx;
	update();
}

void ImagePaint::setImage(const QImage& image) {
	m_images[0] = image;
	imageChanged(m_images[0]);
	m_imgSize = image.size();
	if(!m_imgLoaded)
		initGPU();
	m_imgLoaded = true;
	cleanFBO();
	initFBO();
	update();
}

void ImagePaint::setOriginX(double value) {
	double x = value;
	if (m_moveOrigin.x() != x) {
		m_moveOrigin.setX(x);
		originXChanged(x);
		update();
	}
}

void ImagePaint::setOriginY(double value) {
	double y = value;
	if (m_moveOrigin.y() != y) {
		m_moveOrigin.setY(y);
		originYChanged(y);
		update();
	}
}

void ImagePaint::setZoom(double value) {
	if (value != m_zoom) {
		m_zoom = value;
		zoomChanged(m_zoom);
		update();
	}
}

void ImagePaint::setResolution(double value) {
	if (value != m_resolution) {
		m_resolution = value;
		resolutionChanged(m_resolution);
		update();
	}
}

void ImagePaint::zoomIn() {
	setZoom(m_zoom * 2);
}

void ImagePaint::zoomOut() {
	setZoom(m_zoom / 2);
}

void ImagePaint::reset() {
	m_zoom = 1.0;
	m_resolution = 1.0;
	m_moveOrigin = QPoint(0, 0);
	m_zoomOrigin = QPoint(0, 0);
	m_moveOriginKeep = QPoint(0, 0);
	m_zoomOriginKeep = QPoint(0, 0);
	zoomChanged(1);
	resolutionChanged(1);
	originXChanged(0);
	originYChanged(0);
	update();
	MainWindow::getInstance()->setStatusMouse(QString("| Pos: (%1,%2)").arg(m_moveOrigin.x()).arg(m_moveOrigin.y()));
	MainWindow::getInstance()->setStatusZoom(QString("| Zoom: %1%").arg(m_zoom*100.0));
	MainWindow::getInstance()->setStatusResolution(QString("| Resl: %1%").arg(m_resolution*100.0));
}

void ImagePaint::hold() {
	m_images[1] = m_images[0];
	update();
}

void ImagePaint::toggle() {
	m_index = 1 - m_index;
	QString title = window()->windowTitle().replace(" -- hold", "");
	if (m_index)
		title += " -- hold";
	window()->setWindowTitle(title);
	update();
} 

void ImagePaint::copy() {
	QClipboard *clipboard = QApplication::clipboard();
	clipboard->setImage(image());
}

void ImagePaint::savePNG(const QString& text, bool bauto) {
	QSettings settings;
	QString inputPath = window()->windowFilePath();
	QString outputPath = settings.value("savename", inputPath).toString();
	
	QString filename;
	if(!bauto)
	{
		QFileInfo fi(inputPath);
		QFileInfo fo(outputPath);
		if (!fi.baseName().isEmpty()) {
			QFileInfo fn(fo.dir(), fi.baseName() + "-out.png");
			filename  = fn.absoluteFilePath();
		} else {
			filename  = fo.absolutePath();
		}

		filename = QFileDialog::getSaveFileName(this, "Save PNG", filename, 
			"PNG Format (*.png);;All files (*.*)");
	}
	else
		filename = text;

	if (!filename.isEmpty()) {
		//QImage tmp(image());
		_tex[T_FIN]->bind();
		unsigned char *pData = new unsigned char[m_imgSize.width()*m_imgSize.height()*3];
		glGetTexImage(GL_TEXTURE_2D,0,GL_RGB,GL_UNSIGNED_BYTE,pData);
		QImage tmp = QImage(pData, m_imgSize.width(), m_imgSize.height(), QImage::Format_RGB888);

		if (!text.isEmpty()) tmp.setText("Description", text);
		if (!tmp.save(filename)) {
			QMessageBox::critical(this, "Error", QString("Saving PNG '%1' failed!").arg(filename));
			return;
		}
		settings.setValue("savename", filename);
	}
}

bool ImagePaint::eventFilter( QObject *watched, QEvent *e ) {
	if (!hasFocus() && (m_mode == MODE_MOVE) && (e->type() == QEvent::KeyPress)) {
		QKeyEvent *k = (QKeyEvent*)e;
		if (k->key() == Qt::Key_Space) {
			QPoint gp = mapFromGlobal(QCursor::pos());
			if (rect().contains(gp)) {
				setFocus(Qt::OtherFocusReason);
				keyPressEvent(k);
				return true;
			}
		}
	}
	return QWidget::eventFilter(watched, e);
}


void ImagePaint::leaveEvent( QEvent *e ) {
	//if (m_spacePressed) {
	//	setCursor(m_cursor);
	//	m_cursor = QCursor();
	//	m_spacePressed = false;
	//}
	//if (m_overlay) {
	//	m_overlay->setVisible(false);
	//}
	QWidget::leaveEvent(e);
}

void ImagePaint::setCurrentTool(int idx) {
	m_toolMode = idx;
	switch (idx)
	{
	case E_MOVE:
		setCursor(Qt::OpenHandCursor);
		break;
	case E_ROTO:
		setCursor(Qt::ArrowCursor);
		break;
	default:
		setCursor(Qt::ArrowCursor);
	}
}

void ImagePaint::keyPressEvent( QKeyEvent *e ) {
	if ((m_mode == MODE_MOVE) && (e->key() == Qt::Key_Space)) {
		if (!m_spacePressed) {
			if (m_overlay && m_overlay->isVisible()) m_overlay->setVisible(false);
			m_cursor = cursor();
			setCursor(Qt::OpenHandCursor);
			m_spacePressed = true;
		}
	} else {
		if(e->key()==Qt::Key_Control)
			m_toolMode = E_ZOOM;
		QWidget::keyPressEvent(e);
	}
}


void ImagePaint::keyReleaseEvent( QKeyEvent *e ) {
	if (e->key() == Qt::Key_Space) {
		if (!e->isAutoRepeat()) {
			if (m_mode == MODE_MOVE) {
				setCursor(m_cursor);
				m_cursor = QCursor();
			}
			m_spacePressed = false;
		}
	} else {
		if(e->key()==Qt::Key_Control)
			m_toolMode = E_MOVE;
		QWidget::keyReleaseEvent(e);
	}
}


void ImagePaint::mousePressEvent( QMouseEvent *e ) {
	m_dragButton = e->button();
	if(m_toolMode == E_MOVE )
	{
		if(m_dragButton == Qt::LeftButton || m_dragButton == Qt::RightButton)
		{
			setCursor(Qt::ClosedHandCursor);
			m_moveStart =  e->pos();
			update();
		}
	}
	else if(m_toolMode == E_ZOOM )
	{
		if(m_dragButton == Qt::LeftButton || m_dragButton == Qt::RightButton)
		{
			setCursor(Qt::ClosedHandCursor);
			m_moveStart = e->pos();
			update();
		}
	}
	else if(m_toolMode == E_ROTO && m_dragButton == Qt::LeftButton)
	{
		m_currPosWnd = e->pos();
		m_currPosImg = view2image(m_currPosWnd);
		update();
	}
//	if ((m_mode == MODE_MOVE) && (e->button() != Qt::NoButton)) 
//	{
//		if (!m_spacePressed || (e->buttons() != Qt::LeftButton)) 
//		{
////			m_mode = MODE_DRAG;
//			m_dragStart =  e->pos();
//			m_dragButton = e->button();
////			mouseDragBegin(e);
//			if (e->isAccepted()) return;
//		} 
//		setCursor(Qt::ClosedHandCursor);
//		m_mode = MODE_ROTO;
//		m_dragStart = e->pos();
//		m_dragButton = e->button();
//		m_dragOrigin = m_origin;
//		update();
//	}
	m_currPosWnd = e->pos();
	m_currPosImg = view2image(m_currPosWnd);
	update();
}


void ImagePaint::mouseMoveEvent( QMouseEvent *e ) {
	if(m_toolMode == E_MOVE )
	{
		if(m_dragButton == Qt::LeftButton)
		{
			m_moveOrigin = m_moveOriginKeep + QPointF(e->pos() - m_moveStart) / m_zoom;
			originXChanged(m_moveOrigin.x());
			originYChanged(m_moveOrigin.y());
			update();
		}
		else if(m_dragButton == Qt::RightButton)
		{
			double u = double(e->pos().y()-m_moveStart.y())/double(m_wndSize.height());
			setZoom(m_zoom * (1 + u));
			m_moveStart = e->pos();
		}
	}
	else if(m_toolMode == E_ZOOM )
	{
		if(m_dragButton==Qt::LeftButton)
		{
			m_zoomOrigin = m_zoomOriginKeep + QPointF(e->pos() - m_moveStart) / (m_resolution*m_zoom);
			update();
		}
		else if(e->pos().y()!=m_moveStart.y() && m_dragButton == Qt::RightButton)
		{
			double u = double(e->pos().y()-m_moveStart.y())/double(m_imgSize.height());
			setResolution(m_resolution * (1 + u));
			m_moveStart = e->pos();
		}
	}
	else if(m_toolMode == E_ROTO && m_dragButton == Qt::LeftButton)
	{
		m_currPosWnd = e->pos();
		m_currPosImg = view2image(m_currPosWnd);
		//m_currPosImg = QPointF(338.0,401.0);
		update();
	}
	//if (m_mode == MODE_ROTO) 
	//{
	//	m_origin = m_dragOrigin + QPointF(e->pos() - m_dragStart) / m_zoom / m_resolution;
	//	originXChanged(m_origin.x());
	//	originYChanged(m_origin.y());
	//	update();
	//} 
	//if (m_overlay && !m_spacePressed && (m_mode != MODE_ROTO)) 
	//{
	//	QSize sz = m_overlay->size();
	//	m_overlay->move(e->pos().x() - sz.width()/2, e->pos().y() - sz.height()/2);
	//	if (m_overlay->isHidden())
	//		m_overlay->setVisible(true);
	//}
	m_currPosWnd = e->pos();
	m_currPosImg = view2image(m_currPosWnd);
	update();
	MainWindow::getInstance()->setStatusMouse(QString("| Pos: (%1,%2)").arg(m_moveOrigin.x()).arg(m_moveOrigin.y()));
}


void ImagePaint::mouseReleaseEvent( QMouseEvent *e ) {
	m_dragButton = Qt::NoButton; 
	if(m_toolMode == E_MOVE)
	{
		setCursor(Qt::OpenHandCursor);
		m_moveOriginKeep = m_moveOrigin;
	}
	else if(m_toolMode == E_ZOOM)
	{
		setCursor(Qt::OpenHandCursor);
		m_zoomOriginKeep = m_zoomOrigin;
	}
}

void ImagePaint::wheelEvent(QWheelEvent *e) {
	QSize sz = size();
	double u = e->delta() / 120.0 / 4.0;
	if (u < -0.5) u = -0.5;
	if (u > 0.5) u = 0.5;

	if(m_toolMode == E_MOVE)
	{
		setZoom(m_zoom * (1 + u));
		MainWindow::getInstance()->setStatusZoom(QString("| Zoom: %1%").arg(m_zoom*100.0));
	}
	else if(m_toolMode == E_ZOOM)
	{
		setResolution(m_resolution * (1 + u));
		MainWindow::getInstance()->setStatusResolution(QString("| Resl: %1%").arg(m_resolution*100.0));
	}
}

void ImagePaint::draw(QPainter& p, const QRectF& R, const QImage& image) {
	QRect aR = R.toAlignedRect();
	p.drawImage(aR.x(), aR.y(), image.copy(aR));
}
//////////////////////////////////////////////////////////////////////////// opengl
void ImagePaint::initGPU() {
	finalizeGPU();

	glPushAttrib(GL_ALL_ATTRIB_BITS);
	glDisable(GL_LIGHTING);

	initFBO();
	initShaders();
}
void ImagePaint::finalizeGPU() {
	cleanFBO();
	cleanShaders();
	//glPopAttrib();
}

void ImagePaint::cleanFBO() {
	for(int i=0; i<NB_FBO; i++)
	{
		delete _fbo[i];
		_fbo[i] = NULL;
	}
	for(int i=0;i<NB_TEX;++i) 
	{
		delete _tex[i];
		_tex[i] = NULL;
	}
}

void ImagePaint::cleanShaders() {
	for(int i=0;i<NB_PROG;++i) 
	{
		if(_prog[i]!=NULL) 
		{
			delete _prog[i];
			_prog[i] = NULL;
		}
	}
}
void ImagePaint::reloadShaders() {
	if(_prog[0]!=NULL)
	{
		for(int i=0; i<NB_PROG; i++)
			_prog[i]->reload();
		setShaderParameters();
		update();
		std::cout<< "Shaders have been successfully reloaded!" << std::endl;
	}
}

void ImagePaint::initializeGL() {
	/// glew initialization
	makeCurrent();
	GLenum err = glewInit();
	if (GLEW_OK != err)
		fprintf(stderr, "Error: %s\n",glewGetErrorString(err));

	/// GPU initialization
	for(int i=0; i<NB_FBO; i++)
		_fbo[i] = NULL;
	for(int i=0; i<NB_TEX; i++)
		_tex[i] = NULL;
	for(int i=0; i<NB_PROG; i++)
		_prog[i] = NULL;

	if(m_imgLoaded)
		initGPU();

	/// opengl initialization
	glDisable(GL_LIGHTING);
	glDisable(GL_BLEND);
	glDisable(GL_MULTISAMPLE);
	glEnable(GL_DEPTH_TEST);

	QColor bc = palette().color(QPalette::Base);
	glClearColor(float(bc.red())/255.0f,float(bc.green())/255.0f,float(bc.blue())/255.0f,0);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
}

void ImagePaint::loadTexture(int tex_id, const QImage &qimg)
{
	_tex[tex_id]->bind();
	if(qimg.format()!=QImage::Format_RGB32)
		qimg.convertToFormat(QImage::Format_RGB32);

	QImage tmp = qimg;//.mirrored(false,true);
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, qimg.width(), qimg.height(), 0,
		GL_BGRA, GL_UNSIGNED_BYTE, tmp.bits());
}
void ImagePaint::initNoiseTex(int tex_id, const QSize &imgSize)
{
	int num = imgSize.width()*imgSize.height()*4;
	float *pNoise = new float[num];
	int maxsc = 3000;
	for (int i=0; i<num; i++)
	{		
		pNoise[i] = float((rand()%maxsc))/float(maxsc-1);
	}
	_tex[tex_id] = new FloatTexture2D(TextureFormat(GL_TEXTURE_2D,imgSize.width(),imgSize.height(),
		GL_RGBA16F_ARB,GL_RGBA,GL_FLOAT),TextureParams(GL_LINEAR,GL_LINEAR),pNoise);
	delete []pNoise;
}

void ImagePaint::initFBO() {
	if(_fbo[FBO1]==NULL)
	{
		for(int i=0; i<NB_FBO; i++)
			_fbo[i] = new FramebufferObject();

		_tex[T_ORG	] = new FloatTexture2D(TextureFormat(GL_TEXTURE_2D,m_imgSize.width(),m_imgSize.height(),
			GL_RGBA32F_ARB,GL_RGBA,GL_FLOAT), TextureParams(GL_NEAREST,GL_LINEAR));

		_tex[T_CHAN	] = new FloatTexture2D(_tex[T_ORG]->format(),TextureParams(GL_NEAREST,GL_LINEAR));

		_tex[T_GRAD1] = new FloatTexture2D(_tex[T_ORG]->format(),TextureParams(GL_NEAREST,GL_LINEAR));
		_tex[T_GRAD2] = new FloatTexture2D(_tex[T_ORG]->format(),TextureParams(GL_NEAREST,GL_LINEAR));

		_tex[T_MLS_P1	] = new FloatTexture2D(_tex[T_ORG]->format(),TextureParams(GL_NEAREST,GL_NEAREST));
		_tex[T_MLS_P2	] = new FloatTexture2D(_tex[T_ORG]->format(),TextureParams(GL_NEAREST,GL_NEAREST));
		_tex[T_MLS_P3	] = new FloatTexture2D(_tex[T_ORG]->format(),TextureParams(GL_NEAREST,GL_NEAREST));
		_tex[T_MLS_P4	] = new FloatTexture2D(_tex[T_ORG]->format(),TextureParams(GL_NEAREST,GL_NEAREST));

		_tex[T_MLS	] = new FloatTexture2D(_tex[T_ORG]->format(),TextureParams(GL_NEAREST,GL_LINEAR));
		_tex[T_CURV	] = new FloatTexture2D(_tex[T_ORG]->format(),TextureParams(GL_NEAREST,GL_LINEAR));

		_tex[T_FIN	] = new FloatTexture2D(_tex[T_ORG]->format(),TextureParams(GL_NEAREST,GL_LINEAR));
	}

	/// load image to gpu texture	
	loadTexture(T_ORG,m_images[0]);

	initNoiseTex(T_NOISE,m_imgSize);

	/// bind to fbo attachment
	_fbo[FBO1]->bind();
	_fbo[FBO1]->unattachAll();
	{
		_tex[T_CHAN]->bind();
		_fbo[FBO1]->attachTexture(_tex[T_CHAN]->format().target(),_tex[T_CHAN]->id(),GL_COLOR_ATTACHMENT0_EXT);

		_tex[T_GRAD1]->bind();
		_fbo[FBO1]->attachTexture(_tex[T_GRAD1]->format().target(),_tex[T_GRAD1]->id(),GL_COLOR_ATTACHMENT1_EXT);

		_tex[T_GRAD2]->bind();
		_fbo[FBO1]->attachTexture(_tex[T_GRAD2]->format().target(),_tex[T_GRAD2]->id(),GL_COLOR_ATTACHMENT2_EXT);

		_tex[T_MLS_P1]->bind();
		_fbo[FBO1]->attachTexture(_tex[T_MLS_P1]->format().target(),_tex[T_MLS_P1]->id(),GL_COLOR_ATTACHMENT3_EXT);

		_tex[T_MLS_P2]->bind();
		_fbo[FBO1]->attachTexture(_tex[T_MLS_P2]->format().target(),_tex[T_MLS_P2]->id(),GL_COLOR_ATTACHMENT4_EXT);

		_tex[T_MLS_P3]->bind();
		_fbo[FBO1]->attachTexture(_tex[T_MLS_P3]->format().target(),_tex[T_MLS_P3]->id(),GL_COLOR_ATTACHMENT5_EXT);

		_tex[T_MLS_P4]->bind();
		_fbo[FBO1]->attachTexture(_tex[T_MLS_P4]->format().target(),_tex[T_MLS_P4]->id(),GL_COLOR_ATTACHMENT6_EXT);
	}
	_fbo[FBO1]->isValid();

	_fbo[FBO2]->bind();
	_fbo[FBO2]->unattachAll();
	{
		_tex[T_MLS]->bind();
		_fbo[FBO2]->attachTexture(_tex[T_MLS]->format().target(),_tex[T_MLS]->id(),GL_COLOR_ATTACHMENT0_EXT);

		_tex[T_CURV]->bind();
		_fbo[FBO2]->attachTexture(_tex[T_CURV]->format().target(),_tex[T_CURV]->id(),GL_COLOR_ATTACHMENT1_EXT);

		_tex[T_FIN]->bind();
		_fbo[FBO2]->attachTexture(_tex[T_FIN]->format().target(),_tex[T_FIN]->id(),GL_COLOR_ATTACHMENT2_EXT);
	}
	_fbo[FBO2]->isValid();

	FramebufferObject::unbind();
}

void ImagePaint::setShaderParameters() {
	_prog[P_CHAN]->enable();
	_prog[P_CHAN]->addUniform("wh");
	_prog[P_CHAN]->addUniform("rx");
	_prog[P_CHAN]->addUniform("ry");
	_prog[P_CHAN]->addUniform("alpha");
	_prog[P_CHAN]->addUniform("heightType");
	_prog[P_CHAN]->addUniform("txOrg");
	_prog[P_CHAN]->setUniformTexture("txOrg",	0,GL_TEXTURE_2D,_tex[T_ORG]->id());
	_prog[P_CHAN]->disable();

	_prog[P_GRAD]->enable();
	_prog[P_GRAD]->addUniform("wh");
	_prog[P_GRAD]->addUniform("noiseAlp");
	_prog[P_GRAD]->addUniform("txOrg");
	_prog[P_GRAD]->addUniform("txNoise");
	_prog[P_GRAD]->setUniformTexture("txOrg",	0,GL_TEXTURE_2D,_tex[T_CHAN]->id());
	_prog[P_GRAD]->setUniformTexture("txNoise",	1,GL_TEXTURE_2D,_tex[T_NOISE]->id());
	_prog[P_GRAD]->disable();

	_prog[P_MLS_P]->enable();
	_prog[P_MLS_P]->addUniform("wh");
	_prog[P_MLS_P]->addUniform("bSketch");
	_prog[P_MLS_P]->addUniform("diffSize");
	_prog[P_MLS_P]->addUniform("diffType");
	_prog[P_MLS_P]->addUniform("zoomWnd");
	_prog[P_MLS_P]->addUniform("txOrg");
	_prog[P_MLS_P]->addUniform("txDert01");
	_prog[P_MLS_P]->addUniform("txDert02");
	_prog[P_MLS_P]->setUniformTexture("txOrg",		0,GL_TEXTURE_2D,_tex[T_CHAN]->id());
	_prog[P_MLS_P]->setUniformTexture("txDert01",	1,GL_TEXTURE_2D,_tex[T_GRAD1]->id());
	_prog[P_MLS_P]->setUniformTexture("txDert02",	2,GL_TEXTURE_2D,_tex[T_GRAD2]->id());
	_prog[P_MLS_P]->disable();

	_prog[P_MLS]->enable();
	_prog[P_MLS]->addUniform("wh");
	_prog[P_MLS]->addUniform("bSketch");
	_prog[P_MLS]->addUniform("diffSize");
	_prog[P_MLS]->addUniform("diffType");
	_prog[P_MLS]->addUniform("zoomWnd");
	_prog[P_MLS]->addUniform("txOrg");
	_prog[P_MLS]->addUniform("txDert01");
	_prog[P_MLS]->addUniform("txDert02");
	_prog[P_MLS]->addUniform("txMLSP1");
	_prog[P_MLS]->addUniform("txMLSP2");
	_prog[P_MLS]->addUniform("txMLSP3");
	_prog[P_MLS]->addUniform("txMLSP4");
	_prog[P_MLS]->addUniform("txNoise");
	_prog[P_MLS]->setUniformTexture("txOrg",	0,GL_TEXTURE_2D,_tex[T_CHAN]->id());
	_prog[P_MLS]->setUniformTexture("txDert01",	1,GL_TEXTURE_2D,_tex[T_GRAD1]->id());
	_prog[P_MLS]->setUniformTexture("txDert02",	2,GL_TEXTURE_2D,_tex[T_GRAD2]->id());
	_prog[P_MLS]->setUniformTexture("txMLSP1",	3,GL_TEXTURE_2D,_tex[T_MLS_P1]->id());
	_prog[P_MLS]->setUniformTexture("txMLSP2",	4,GL_TEXTURE_2D,_tex[T_MLS_P2]->id());
	_prog[P_MLS]->setUniformTexture("txMLSP3",	5,GL_TEXTURE_2D,_tex[T_MLS_P3]->id());
	_prog[P_MLS]->setUniformTexture("txMLSP4",	6,GL_TEXTURE_2D,_tex[T_MLS_P4]->id());
	_prog[P_MLS]->setUniformTexture("txNoise",	7,GL_TEXTURE_2D,_tex[T_NOISE]->id());
	_prog[P_MLS]->disable();

	_prog[P_FIN]->enable();
	_prog[P_FIN]->addUniform("wh");
	_prog[P_FIN]->addUniform("display");
	_prog[P_FIN]->addUniform("zoomWnd");
	_prog[P_FIN]->addUniform("txOrg");
	_prog[P_FIN]->addUniform("txDert01");
	_prog[P_FIN]->addUniform("txDert02");
	_prog[P_FIN]->addUniform("txFlow");
	_prog[P_FIN]->addUniform("txCurv");
	_prog[P_FIN]->addUniform("txNoise");
	_prog[P_FIN]->setUniformTexture("txOrg",	0,GL_TEXTURE_2D,_tex[T_CHAN]->id());
	_prog[P_FIN]->setUniformTexture("txDert01",	1,GL_TEXTURE_2D,_tex[T_GRAD1]->id());
	_prog[P_FIN]->setUniformTexture("txDert02",	2,GL_TEXTURE_2D,_tex[T_GRAD2]->id());
	_prog[P_FIN]->setUniformTexture("txFlow",	3,GL_TEXTURE_2D,_tex[T_MLS]->id());
	_prog[P_FIN]->setUniformTexture("txCurv",	4,GL_TEXTURE_2D,_tex[T_CURV]->id());
	_prog[P_FIN]->setUniformTexture("txNoise",	5,GL_TEXTURE_2D,_tex[T_NOISE]->id());
	_prog[P_FIN]->disable();
}
void ImagePaint::initShaders() {
	static std::string SHADER_DIR = "../mlsfield/shaders/";
	_prog[P_CHAN	]= new GPUProgram(SHADER_DIR+"default.vs",		SHADER_DIR+"channel.fs");
	_prog[P_GRAD	]= new GPUProgram(SHADER_DIR+"default.vs",		SHADER_DIR+"gradient.fs");
	_prog[P_MLS_P	]= new GPUProgram(SHADER_DIR+"default.vs",		SHADER_DIR+"mlsPre.fs");
	_prog[P_MLS		]= new GPUProgram(SHADER_DIR+"default.vs",		SHADER_DIR+"mls.fs");
	_prog[P_FIN		]= new GPUProgram(SHADER_DIR+"default.vs",		SHADER_DIR+"fin.fs");

	setShaderParameters();
}

void ImagePaint::resizeGL(int w, int h) {
	m_wndSize = QSize(w,h);
}

void ImagePaint::paintGL() {
	if(!m_imgLoaded)
		return;

	QRectF zoomWnd = QRectF(-m_zoomOrigin
		+QPointF(m_imgSize.width(),m_imgSize.height())*(0.5-0.5/float(m_resolution)),
		QSizeF(m_imgSize)/float(m_resolution));
	float ws = 1.0f/float(m_imgSize.width());
	float hs = 1.0f/float(m_imgSize.height());

	glDisable(GL_LIGHTING);
	glColor4f(1,1,1,1);
	glEnable(GL_TEXTURE_2D);

	/////////////////////////////////////////////// render to texture
	_fbo[FBO1]->bind();
	swapToLocalMode(m_imgSize.width(), m_imgSize.height());

	glDrawBuffer(*FramebufferObject::buffers(0));
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	_prog[P_CHAN]->enable();
	glEnable(GL_TEXTURE_2D);
	_prog[P_CHAN]->setUniform2f("wh",ws,hs);
	_prog[P_CHAN]->setUniform1f("rx",float(m_rx));
	_prog[P_CHAN]->setUniform1f("ry",float(m_ry));
	_prog[P_CHAN]->setUniform1f("alpha",float(m_alpha));
	_prog[P_CHAN]->setUniform1i("heightType",m_heightType);
	drawQuad(m_imgSize.width(), m_imgSize.height());
	_prog[P_CHAN]->disable();

	glDrawBuffers(2,FramebufferObject::buffers(1));
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	_prog[P_GRAD]->enable();
	glEnable(GL_TEXTURE_2D);
	_prog[P_GRAD]->setUniform2f("wh",ws,hs);
	_prog[P_GRAD]->setUniform1f("noiseAlp",float(m_noiseAlp));
	_prog[P_GRAD]->setUniform1i("flowType",m_flowType);
	drawQuad(m_imgSize.width(), m_imgSize.height());
	_prog[P_GRAD]->disable();

	if(m_isCPU && m_cpuRecompute)	/// reconstruct on CPU
	{
		if(m_diffType==0)
			std::cout << "Computing MLS with isotropic linear!" << std::endl;
		else if(m_diffType==1)
			std::cout << "Computing MLS with constant!" << std::endl;
		else if(m_diffType==2)
			std::cout << "Computing MLS with full linear!" << std::endl;
		else 
			std::cout << "Flow type is out of range!" << std::endl;
		
		computeFullLinear();
		_fbo[FBO2]->bind();
	}
	else if(!m_isCPU)	/// reconstruct on GPU
	{
		glDrawBuffers(4,FramebufferObject::buffers(3));
		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
		_prog[P_MLS_P]->enable();
		glEnable(GL_TEXTURE_2D);
		_prog[P_MLS_P]->setUniform2f("wh",ws,hs);
		_prog[P_MLS_P]->setUniform1i("diffSize",m_diffSize);
		_prog[P_MLS_P]->setUniform1i("diffType",m_diffType);
		_prog[P_MLS_P]->setUniform1i("bSketch",0);
		_prog[P_MLS_P]->setUniform4f("zoomWnd",zoomWnd.x(),zoomWnd.y(),zoomWnd.width(),zoomWnd.height());
		drawQuad(m_imgSize.width(), m_imgSize.height());
		_prog[P_MLS_P]->disable();

		_fbo[FBO2]->bind();

		glDrawBuffers(2,FramebufferObject::buffers(0));
		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
		_prog[P_MLS]->enable();
		glEnable(GL_TEXTURE_2D);
		_prog[P_MLS]->setUniform2f("wh",ws,hs);
		_prog[P_MLS]->setUniform1i("diffSize",m_diffSize);
		_prog[P_MLS]->setUniform1i("diffType",m_diffType);
		_prog[P_MLS]->setUniform1i("bSketch",0);
		_prog[P_MLS]->setUniform4f("zoomWnd",zoomWnd.x(),zoomWnd.y(),zoomWnd.width(),zoomWnd.height());
		drawQuad(m_imgSize.width(), m_imgSize.height());
		_prog[P_MLS]->disable();
	}
	else	/// using the results computed by CPU
		_fbo[FBO2]->bind();
	
	
	glDrawBuffer(*FramebufferObject::buffers(2));
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	_prog[P_FIN]->enable();
	glEnable(GL_TEXTURE_2D);
	_prog[P_FIN]->setUniform2f("wh",ws,hs);
	_prog[P_FIN]->setUniform1i("flowType",m_flowType);
	_prog[P_FIN]->setUniform1i("display",m_index);
	_prog[P_FIN]->setUniform4f("zoomWnd",zoomWnd.x(),zoomWnd.y(),zoomWnd.width(),zoomWnd.height());
	drawQuad(m_imgSize.width(), m_imgSize.height());
	_prog[P_FIN]->disable();

	swapToWorldMode();

	/////////////////////////////////// render to screen  horizontal flip 
	swapToLocalMode(m_wndSize.width(), m_wndSize.height());
	FramebufferObject::unbind();

	glEnable(GL_TEXTURE_2D);
	QColor bc = palette().color(QPalette::Window);
	glClearColor(float(bc.red())/255.0f,float(bc.green())/255.0f,float(bc.blue())/255.0f,0);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	
	_tex[T_FIN]->bind();
	QSize isz = m_imgSize*m_zoom;
	drawQuadScreen(isz.width(), isz.height(), 
		float(m_wndSize.width()-isz.width())/2.0f + m_moveOrigin.x()*m_zoom, 
		float(m_wndSize.height()-isz.height())/2.0f - m_moveOrigin.y()*m_zoom);
	glClearColor(1,1,1,0);

	if(m_showFittingCircle)
		drawFittingCircle();

//	if(m_index==2)
//		drawArrowWindow(8);

	//glDisable(GL_TEXTURE_2D);
	//glLineWidth(2.0);
	//glBegin(GL_LINES);
	//glColor3f(0,0,0);
	//float yy = float(m_wndSize.height())/2.0 + m_alpha*m_imgSize.height();
	//glVertex3f(100,yy,0.0001);
	//glVertex3f(m_wndSize.width()-100,yy,0.0001);
	//glEnd();
	//glColor3f(1,1,1);
	//glEnable(GL_TEXTURE_2D);

	swapToWorldMode();
	m_cpuRecompute = false;
}
void ImagePaint::saveHeightMap(const std::string &filename)
{
	int cn =4;
	const GLenum format1 = GL_RGBA;
	_tex[T_CHAN]->bind();
	float *pData = new float[m_imgSize.width()*m_imgSize.height()*cn];
	glGetTexImage(GL_TEXTURE_2D,0,format1,GL_FLOAT,pData);

	std::ofstream out(filename.c_str());

	out << "#height map object" << std::endl;
	out << "#vertex data" << std::endl;
	for(int yi=0; yi<m_imgSize.height(); yi++)
	{
		int yidx = yi*m_imgSize.width()*cn;
		for(int xi=0; xi<m_imgSize.width(); xi++)
		{
			int xidx = yidx + xi*cn;
			out << "v " << float(xi) << " "
				<< float(yi) << " "	<< pData[xidx] << std::endl;
		}
	}
	out << "#" << m_imgSize.width()*m_imgSize.height() << " vertices" << std::endl;
	delete [] pData;

	out << std::endl;
	out << "#mesh face data" << std::endl;
	for(int yi=0; yi<m_imgSize.height()-1; yi++)
	{
		int yidx = yi*m_imgSize.width();
		for(int xi=0; xi<m_imgSize.width()-1; xi++)
		{
			int xidx = yidx + xi+1;
			out << "f " << xidx << " "<< xidx+1 
				<< " "<< xidx+m_imgSize.width() << " "<< std::endl;
			out << "f " << xidx+1 << " "<< xidx+1+m_imgSize.width() 
				<< " "<< xidx+m_imgSize.width() << " "<< std::endl;
		}
	}
	out << "#" << 2*(m_imgSize.width()-1)*(m_imgSize.height()-1) << " faces" << std::endl;
	out.close();

	std::cout << "Height map object has been saved!" << std::endl;
}
void ImagePaint::drawArrowWindow(int sz)
{
	std::vector<float>	vTruth(sz*sz*2);
	std::vector<float>	vNoise(sz*sz*2);
	std::vector<float>	vEstim(sz*sz*2);

	int cn =4;
	const GLenum format1 = GL_RGBA;
	_tex[T_GRAD1]->bind();
	float *pGrad1 = new float[m_imgSize.width()*m_imgSize.height()*cn];
	glGetTexImage(GL_TEXTURE_2D,0,format1,GL_FLOAT,pGrad1);

	_tex[T_GRAD2]->bind();
	float *pGrad2 = new float[m_imgSize.width()*m_imgSize.height()*cn];
	glGetTexImage(GL_TEXTURE_2D,0,format1,GL_FLOAT,pGrad2);

	_tex[T_MLS]->bind();
	float *pMLS = new float[m_imgSize.width()*m_imgSize.height()*cn];
	glGetTexImage(GL_TEXTURE_2D,0,format1,GL_FLOAT,pMLS);

	int mcen = m_imgSize.width()/2;
	int yoffset = m_alpha*m_imgSize.height();
	int midx = (mcen-yoffset-sz/2)*m_imgSize.width()*cn+(mcen-sz/2)*cn;
	float err = 0.0;
	for(int yi=0; yi<sz; yi++)
	{
		int yidx = yi*sz*2;
		for(int xi=0; xi<sz; xi++)
		{
			int sidx = midx + yi*m_imgSize.width()*cn + xi*cn;
			vTruth[yidx+xi*2] = -pGrad2[sidx+3];
			vTruth[yidx+xi*2+1] = pGrad1[sidx+3];
			vNoise[yidx+xi*2] = -pGrad2[sidx];
			vNoise[yidx+xi*2+1] = pGrad1[sidx];
			vEstim[yidx+xi*2] = -pMLS[sidx+1];
			vEstim[yidx+xi*2+1] = pMLS[sidx];

			float realAngle = atan2(pGrad1[sidx+3], -pGrad2[sidx+3] );
			float estiAngle = atan2(pMLS[sidx], -pMLS[sidx+1] );
			float pi = 3.1415926;
			err += min(min(abs(realAngle-estiAngle),abs(realAngle+pi-estiAngle)), abs(realAngle-pi-estiAngle))*2.0/pi;
		}
	}
	delete []pGrad1;
	delete []pGrad2;
	delete []pMLS;
	std::cout << "reconstruct errors = " << err << std::endl;

	int sc = int(m_imgSize.width()/sz);
	float dep = 0.001f;
	glDisable(GL_TEXTURE_2D);
	glLineWidth(2.0);
	glBegin(GL_LINES);
	for(int yi=0; yi<sz; yi++)
	{
		int yidx = yi*sz*2;
		for(int xi=0; xi<sz; xi++)
		{
			/// in software window
			float xcen = float(m_wndSize.width()-m_imgSize.width())/2.0f + float(xi*sc) + float(sc)/2.0f;
			float ycen = float(m_wndSize.height()-m_imgSize.height())/2.0f + float(yi*sc) + float(sc)/2.0f;
			/// in image window
		//	float xcen = float(xi*sc) + float(sc)/2.0f;
		//	float ycen = float(yi*sc) + float(sc)/2.0f;
			/// ground truth
			float xbgnT = xcen - float(sc)*vTruth[yidx+xi*2]/2.0f;
			float ybgnT = ycen - float(sc)*vTruth[yidx+xi*2+1]/2.0f;
			float xendT = xcen + float(sc)*vTruth[yidx+xi*2]/2.0f;
			float yendT = ycen + float(sc)*vTruth[yidx+xi*2+1]/2.0f;
			glColor3f(0,0,0);
			glVertex3f(xbgnT,ybgnT,dep);
			glVertex3f(xendT,yendT,dep);

			/// input data with noises
			float xbgnN = xcen - float(sc)*vNoise[yidx+xi*2]/2.0f;
			float ybgnN = ycen - float(sc)*vNoise[yidx+xi*2+1]/2.0f;
			float xendN = xcen + float(sc)*vNoise[yidx+xi*2]/2.0f;
			float yendN = ycen + float(sc)*vNoise[yidx+xi*2+1]/2.0f;
			glColor3f(0.65,0.65,0.65);
			glVertex3f(xbgnN,ybgnN,dep);
			glVertex3f(xendN,yendN,dep);

			/// estimated
			float xbgnE = xcen - float(sc)*vEstim[yidx+xi*2]/2.0f;
			float ybgnE = ycen - float(sc)*vEstim[yidx+xi*2+1]/2.0f;
			float xendE = xcen + float(sc)*vEstim[yidx+xi*2]/2.0f;
			float yendE = ycen + float(sc)*vEstim[yidx+xi*2+1]/2.0f;
			glColor3f(1,0,0);
			glVertex3f(xbgnE,ybgnE,dep);
			glVertex3f(xendE,yendE,dep);
		}
	}
	glEnd();
	glColor3f(1,1,1);
	glEnable(GL_TEXTURE_2D);
}

void ImagePaint::setComputeFullLinear()
{
	m_cpuRecompute = true;
	paintGL();	
}
void ImagePaint::setCPUUsing(bool isUsing)
{
	m_isCPU = isUsing;
	if(isUsing==false)
		paintGL();
}